use std::sync::Arc;

use vulkano::{
    device::Device,
    pipeline::{
        graphics::{
            color_blend::{ColorBlendAttachmentState, ColorBlendState},
            depth_stencil::{DepthState, DepthStencilState},
            input_assembly::InputAssemblyState,
            multisample::MultisampleState,
            rasterization::RasterizationState,
            vertex_input::{Vertex, VertexDefinition},
            viewport::{Viewport, ViewportState},
            GraphicsPipelineCreateInfo,
        },
        layout::PipelineDescriptorSetLayoutCreateInfo,
        GraphicsPipeline, PipelineLayout, PipelineShaderStageCreateInfo,
    },
    render_pass::{RenderPass, Subpass},
    shader::ShaderModule,
};

use super::vertex::{Normal, Position};

pub struct PipelineInfo {
    pub vs: Arc<ShaderModule>,
    pub fs: Arc<ShaderModule>,
}

pub fn make_pipeline(
    device: Arc<Device>,
    pipeline_info: &PipelineInfo,
    pass_info: &PassInfo,
) -> Arc<GraphicsPipeline> {
    let vs = pipeline_info.vs.entry_point("main").unwrap();
    let fs = pipeline_info.fs.entry_point("main").unwrap();

    let vertex_input_state = [Position::per_vertex(), Normal::per_vertex()]
        .definition(&vs)
        .unwrap();
    let stages = [
        PipelineShaderStageCreateInfo::new(vs),
        PipelineShaderStageCreateInfo::new(fs),
    ];
    let layout = PipelineLayout::new(
        device.clone(),
        PipelineDescriptorSetLayoutCreateInfo::from_stages(&stages)
            .into_pipeline_layout_create_info(device.clone())
            .unwrap(),
    )
    .unwrap();
    let subpass = pass_info.subpass().unwrap();

    let extent = pass_info.extent;

    GraphicsPipeline::new(
        device,
        None,
        GraphicsPipelineCreateInfo {
            stages: stages.into_iter().collect(),
            vertex_input_state: Some(vertex_input_state),
            input_assembly_state: Some(InputAssemblyState::default()),
            viewport_state: Some(ViewportState {
                viewports: [Viewport {
                    offset: [0.0, 0.0],
                    extent: [extent[0] as f32, extent[1] as f32],
                    depth_range: 0.0..=1.0,
                }]
                .into_iter()
                .collect(),
                ..Default::default()
            }),
            rasterization_state: Some(RasterizationState::default()),
            depth_stencil_state: Some(DepthStencilState {
                depth: Some(DepthState::simple()),
                ..Default::default()
            }),
            multisample_state: Some(MultisampleState::default()),
            color_blend_state: Some(ColorBlendState::with_attachment_states(
                subpass.num_color_attachments(),
                ColorBlendAttachmentState::default(),
            )),
            subpass: Some(subpass.into()),
            ..GraphicsPipelineCreateInfo::layout(layout)
        },
    )
    .unwrap()
}

pub struct PassInfo {
    pub render_pass: Arc<RenderPass>,
    pub subpass_id: u32,
    pub extent: [u32; 3],
}

impl PassInfo {
    pub fn new(render_pass: Arc<RenderPass>, subpass_id: u32, extent: [u32; 3]) -> Self {
        Self {
            render_pass: render_pass.clone(),
            subpass_id,
            extent,
        }
    }

    pub fn subpass(&self) -> Option<Subpass> {
        Subpass::from(self.render_pass.clone(), self.subpass_id)
    }
}
